class MissingMandatoryValue(Exception):
    """Thrown when a variable flagged with '???' value is accessed to
    indicate that the value was not set"""


class ValidationError(Exception):
    """
    Thrown when a value fails validation
    """


class ReadonlyConfigError(Exception):
    """
    Thrown when someone tries to modify a frozen config
    """



import copy
import itertools

import six

from .config import Config, isint
from .errors import ReadonlyConfigError
from .nodes import BaseNode, UntypedNode


class ListConfig(Config):
    def __init__(self, content, parent=None):
        super(ListConfig, self).__init__()
        self.__dict__['content'] = []
        self.__dict__['parent'] = parent
        assert isinstance(content, (list, tuple))
        for item in content:
            if isinstance(item, dict) or isinstance(item, (list, tuple)):
                from omegaconf import OmegaConf
                item = OmegaConf.create(item, parent=self)
            self.append(item)

    def __deepcopy__(self, memodict={}):
        res = ListConfig([])
        self._deepcopy_impl(res)
        return res

    def __getattr__(self, key):
        if isinstance(key, str) and isint(key):
            return self.__getitem__(int(key))
        else:
            raise AttributeError()

    # hide content while inspecting in debugger
    def __dir__(self):
        return [str(x) for x in range(0, len(self))]

    def __len__(self):
        return len(self.content)

    def __getitem__(self, index):
        assert isinstance(index, (int, slice))
        if isinstance(index, slice):
            result = []
            for slice_idx in itertools.islice(range(0, len(self)), index.start, index.stop, index.step):
                val = self._resolve_with_default(key=slice_idx, value=self.content[slice_idx], default_value=None)
                result.append(val)
            return result
        else:
            return self._resolve_with_default(key=index, value=self.content[index], default_value=None)

    def _set_at_index(self, index, value):
        if not isinstance(value, Config) and (isinstance(value, dict) or isinstance(value, list)):
            from omegaconf import OmegaConf
            value = OmegaConf.create(value, parent=self)

        if not Config.is_primitive_type(value):
            full_key = self.get_full_key(index)
            raise ValueError("key {}: {} is not a primitive type".format(full_key, type(value).__name__))

        if self._get_flag('freeze'):
            raise ReadonlyConfigError(self.get_full_key(index))

        if not isinstance(value, BaseNode):
            self.__dict__['content'][index].set_value(value)
        else:
            if not isinstance(value, BaseNode):
                value = UntypedNode(value)
            else:
                value = copy.deepcopy(value)
            self.__dict__['content'][index] = value

    def __setitem__(self, index, value):
        assert isinstance(index, int)
        self._set_at_index(index, value)

    def append(self, item):
        try:
            self.__dict__['content'].append(UntypedNode(None))
            self._set_at_index(len(self) - 1, item)
        except Exception:
            del self.__dict__['content'][len(self) - 1]
            raise

    def insert(self, index, item):
        if self._get_flag('freeze'):
            raise ReadonlyConfigError(self.get_full_key(index))
        try:
            self.content.insert(index, UntypedNode(None))
            self._set_at_index(index, item)
        except Exception:
            del self.__dict__['content'][index]
            raise

    if six.PY2:
        def __getslice__(self, start, stop):
            result = []
            for slice_idx in itertools.islice(range(0, len(self)), start, stop, 1):
                val = self._resolve_with_default(key=slice_idx, value=self.content[slice_idx], default_value=None)
                result.append(val)
            return ListConfig(content=result, parent=self.__dict__['parent'])

    def get_node(self, index):
        assert type(index) == int
        return self.content[index]

    def get(self, index, default_value=None):
        assert type(index) == int
        return self._resolve_with_default(key=index, value=self.content[index], default_value=default_value)

    def pop(self, index=-1):
        if self._get_flag('freeze'):
            raise ReadonlyConfigError(self.get_full_key(index))
        return self._resolve_with_default(key=index, value=self.content.pop(index), default_value=None)

    def sort(self, key=None, reverse=False):
        if self._get_flag('freeze'):
            raise ReadonlyConfigError()

        if key is None:
            def key1(x):
                return x.value()
        else:
            def key1(x):
                return key(x.value())
        self.content.sort(key=key1, reverse=reverse)

    def __eq__(self, other):
        if isinstance(other, list):
            return Config._list_eq(self, ListConfig(other))
        if isinstance(other, ListConfig):
            return Config._list_eq(self, other)
        return NotImplemented

    def __ne__(self, other):
        x = self.__eq__(other)
        if x is not NotImplemented:
            return not x
        return NotImplemented

    def __hash__(self):
        # TODO: should actually iterate
        return hash(str(self))

    def __iter__(self):
        class MyItems(object):
            def __init__(self, l):
                self.lst = l
                self.iterator = iter(l)

            def __iter__(self):
                return self

            # Python 3 compatibility
            def __next__(self):
                return self.next()

            def next(self):
                v = next(self.iterator)
                if isinstance(v, BaseNode):
                    v = v.value()
                return v

        return MyItems(self.content)



import copy
import io
import os
import re
import sys
import warnings
from abc import abstractmethod

import six
import yaml
from collections import defaultdict
from .errors import MissingMandatoryValue, ReadonlyConfigError
from .nodes import BaseNode


def isint(s):
    try:
        int(s)
        return True
    except ValueError:
        return False


def get_yaml_loader():
    loader = yaml.SafeLoader
    loader.add_implicit_resolver(
        u'tag:yaml.org,2002:float',
        re.compile(u'''^(?:
         [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
        |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
        |\\.[0-9_]+(?:[eE][-+][0-9]+)?
        |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
        |[-+]?\\.(?:inf|Inf|INF)
        |\\.(?:nan|NaN|NAN))$''', re.X),
        list(u'-+0123456789.'))
    return loader


class Config(object):
    # static fields
    _resolvers = {}
    _resolvers_cache = defaultdict(lambda: defaultdict(dict))

    def __init__(self):
        if type(self) == Config:
            raise NotImplementedError
        # Flags have 3 modes:
        #   unset : inherit from parent, defaults to false in top level config.
        #   set to true: flag is true
        #   set to false: flag is false
        self.__dict__['flags'] = dict(
            # Frozen config cannot be modified
            freeze=None,
            # Struct config throws a KeyError if a non existing field is accessed
            struct=None
        )

    def save(self, f):
        data = self.pretty()
        if isinstance(f, str):
            with io.open(os.path.abspath(f), 'w', encoding='utf-8') as f:
                f.write(six.u(data))
        elif hasattr(f, 'write'):
            f.write(data)
            f.flush()
        else:
            raise TypeError("Unexpected file type")

    def _set_parent(self, parent):
        assert parent is None or isinstance(parent, Config)
        self.__dict__['parent'] = parent

    def _get_root(self):
        root = self.__dict__['parent']
        if root is None:
            return self
        while root.__dict__['parent'] is not None:
            root = root.__dict__['parent']
        return root

    def __del__(self):
        if 'parent' not in self.__dict__ or self.__dict__['parent'] is None:
            # root node, remove resolver cache
            id_ = id(self)
            if id_ in Config._resolvers_cache:
                del Config._resolvers_cache[id_]

    def _set_flag(self, flag, value):
        assert value is None or isinstance(value, bool)
        self.__dict__['flags'][flag] = value

    def _get_flag(self, flag):
        """
        Returns True if this config node flag is set
        A flag is set if node.set_flag(True) was called
        or one if it's parents is flag is set
        :return:
        """
        if flag in self.__dict__['flags'] and self.__dict__['flags'][flag] is not None:
            return self.__dict__['flags'][flag]

        # root leaf, and frozen is not set.
        if self.__dict__['parent'] is None:
            return False
        else:
            # noinspection PyProtectedMember
            return self.__dict__['parent']._get_flag(flag)

    @abstractmethod
    def get(self, key, default_value=None):
        raise NotImplementedError

    @abstractmethod
    def get_node(self, key):
        """
        returns raw node object for this key
        :param key:
        :return:
        """
        raise NotImplementedError

    def _resolve_with_default(self, key, value, default_value=None):
        """returns the value with the specified key, like obj.key and obj['key']"""

        def is_mandatory_missing(val):
            return type(val) == str and val == '???'

        if isinstance(value, BaseNode):
            value = value.value()

        if default_value is not None and (value is None or is_mandatory_missing(value)):
            value = default_value

        if is_mandatory_missing(value):
            raise MissingMandatoryValue(self.get_full_key(key))

        return self._resolve_single(value) if isinstance(value, str) else value

    def get_full_key(self, key):
        from .listconfig import ListConfig
        from .dictconfig import DictConfig
        full_key = ''
        child = None
        parent = self
        while parent is not None:
            assert isinstance(parent, (DictConfig, ListConfig))
            if isinstance(parent, DictConfig):
                if child is None:
                    full_key = "{}".format(key)
                else:
                    for parent_key, v in parent.items():
                        if id(v) == id(child):
                            if isinstance(child, ListConfig):
                                full_key = "{}{}".format(parent_key, full_key)
                            else:
                                full_key = "{}.{}".format(parent_key, full_key)
                            break
            elif isinstance(parent, ListConfig):
                if child is None:
                    full_key = "[{}]".format(key)
                else:
                    for idx, v in enumerate(parent):
                        if id(v) == id(child):
                            if isinstance(child, ListConfig):
                                full_key = "[{}]{}".format(idx, full_key)
                            else:
                                full_key = "[{}].{}".format(idx, full_key)
                            break
            child = parent
            parent = child.__dict__['parent']

        return full_key

    def __str__(self):
        return self.content.__str__()

    def __repr__(self):
        return self.content.__repr__()

    @abstractmethod
    def __eq__(self, other):
        raise NotImplementedError()

    @abstractmethod
    def __ne__(self, other):
        raise NotImplementedError()

    @abstractmethod
    def __hash__(self):
        raise NotImplementedError()

    # Support pickle
    def __getstate__(self):
        return self.__dict__

    # Support pickle
    def __setstate__(self, d):
        self.__dict__.update(d)

    def __delitem__(self, key):
        if self._get_flag('freeze'):
            raise ReadonlyConfigError(self.get_full_key(key))
        self.content.__delitem__(key)

    def __len__(self):
        return self.content.__len__()

    @abstractmethod
    def __iter__(self):
        raise NotImplementedError()

    def __contains__(self, item):
        return self.content.__contains__(item)

    @staticmethod
    def _select_one(c, key_):
        from .listconfig import ListConfig
        from .dictconfig import DictConfig
        assert isinstance(c, (DictConfig, ListConfig))
        if isinstance(c, DictConfig):
            if key_ in c:
                return c[key_], key_
            else:
                return None, key_
        elif isinstance(c, ListConfig):
            if not isint(key_):
                raise TypeError("Index {} is not an int".format(key_))
            idx = int(key_)
            if idx < 0 or idx + 1 > len(c):
                return None, idx
            return c[idx], idx

    def merge_with_cli(self):
        args_list = sys.argv[1:]
        self.merge_with_dotlist(args_list)

    def merge_with_dotlist(self, dotlist):
        for arg in dotlist:
            args = arg.split('=')
            key = args[0]
            value = None
            if len(args) > 1:
                # load with yaml to get correct automatic typing with the same rules as yaml parsing
                value = yaml.load(args[1], Loader=get_yaml_loader())
            self.update(key, value)

    def update(self, key, value=None):
        from .listconfig import ListConfig
        from .dictconfig import DictConfig
        """Updates a dot separated key sequence to a value"""
        split = key.split('.')
        root = self
        for i in range(len(split) - 1):
            k = split[i]
            # if next_root is a primitive (string, int etc) replace it with an empty map
            next_root, key_ = Config._select_one(root, k)
            if not isinstance(next_root, Config):
                root[key_] = {}
            root = root[key_]

        last = split[-1]

        assert isinstance(root, (DictConfig, ListConfig))
        if isinstance(root, DictConfig):
            root[last] = value
        elif isinstance(root, ListConfig):
            idx = int(last)
            root[idx] = value

    def select(self, key):
        """
        Select a value using dot separated key sequence
        :param key:
        :return:
        """

        split = key.split('.')
        root = self
        for i in range(len(split) - 1):
            if root is None:
                break
            k = split[i]
            root, _ = Config._select_one(root, k)

        if root is None:
            return None

        value, _ = Config._select_one(root, split[-1])
        return value

    def is_empty(self):
        """return true if config is empty"""
        return len(self.content) == 0

    @staticmethod
    def _to_content(conf, resolve):
        from .listconfig import ListConfig
        from .dictconfig import DictConfig
        assert isinstance(conf, Config)
        if isinstance(conf, DictConfig):
            ret = {}
            for key, value in conf.items(resolve=resolve):
                if isinstance(value, Config):
                    ret[key] = Config._to_content(value, resolve)
                else:
                    ret[key] = value
            return ret
        elif isinstance(conf, ListConfig):
            ret = []
            for index, item in enumerate(conf):
                if resolve:
                    item = conf[index]
                if isinstance(item, Config):
                    item = Config._to_content(item, resolve)
                ret.append(item)
            return ret

    def to_container(self, resolve=False):
        return Config._to_content(self, resolve)

    def pretty(self, resolve=False):
        """
        returns a yaml dump of this config object.
        :param resolve: if True, will return a string with the interpolations resolved, otherwise
        interpolations are preserved
        :return: A string containing the yaml representation.
        """
        return yaml.dump(self.to_container(resolve=resolve), default_flow_style=False)

    @staticmethod
    def _map_merge(dest, src):
        """merge src into dest and return a new copy, does not modified input"""
        from .dictconfig import DictConfig

        assert isinstance(dest, DictConfig)
        assert isinstance(src, DictConfig)
        src = copy.deepcopy(src)

        for key, value in src.items(resolve=False):
            if key in dest:
                dest_node = dest.get_node(key)
                if isinstance(dest_node, Config):
                    if isinstance(value, Config):
                        dest_node.merge_with(value)
                    else:
                        dest.__setitem__(key, value)
                else:
                    if isinstance(value, Config):
                        dest.__setitem__(key, value)
                    else:
                        dest_node.set_value(value)
            else:
                dest[key] = src.get_node(key)

    def merge_from(self, *others):
        warnings.warn("Use Config.merge_with() (since 1.1.10)", DeprecationWarning,
                      stacklevel=2)

        self.merge_with(*others)

    def _deepcopy_impl(self, res, _memodict={}):
        # memodict is intentionally not used.
        # Using it can cause python to return objects that were since modified, undoing their modifications!
        res.__dict__['content'] = copy.deepcopy(self.__dict__['content'])
        res.__dict__['flags'] = copy.deepcopy(self.__dict__['flags'])
        # intentionally not deepcopying the parent. this can cause all sorts of mayhem and stack overflow.
        # instead of just re-parent the result node. this will break interpolation in cases of deepcopying
        # a node that is not the root node, but that is almost guaranteed to break anyway.
        Config._re_parent(res)

    @staticmethod
    def _re_parent(node):
        from .listconfig import ListConfig
        from .dictconfig import DictConfig

        # update parents of first level Config nodes to self
        assert isinstance(node, (DictConfig, ListConfig))
        if isinstance(node, DictConfig):
            for _key, value in node.items(resolve=False):
                if isinstance(value, Config):
                    value._set_parent(node)
                    Config._re_parent(value)
        elif isinstance(node, ListConfig):
            for item in node:
                if isinstance(item, Config):
                    item._set_parent(node)
                    Config._re_parent(item)

    def merge_with(self, *others):
        from .listconfig import ListConfig
        from .dictconfig import DictConfig
        """merge a list of other Config objects into this one, overriding as needed"""
        for other in others:
            if other is None:
                raise ValueError("Cannot merge with a None config")
            if isinstance(self, DictConfig) and isinstance(other, DictConfig):
                Config._map_merge(self, other)
            elif isinstance(self, ListConfig) and isinstance(other, ListConfig):
                self.__dict__['content'] = copy.deepcopy(other.content)
            else:
                raise TypeError("Merging DictConfig with ListConfig is not supported")

        # recursively correct the parent hierarchy after the merge
        Config._re_parent(self)

    @staticmethod
    def _resolve_value(root_node, inter_type, inter_key):
        from omegaconf import OmegaConf
        inter_type = ('str:' if inter_type is None else inter_type)[0:-1]
        if inter_type == 'str':
            ret = root_node.select(inter_key)
            if ret is None:
                raise KeyError("{} interpolation key '{}' not found".format(inter_type, inter_key))
        else:
            resolver = OmegaConf.get_resolver(inter_type)
            if resolver is not None:
                ret = resolver(root_node, inter_key)
            else:
                raise ValueError("Unsupported interpolation type {}".format(inter_type))

        return ret

    def _resolve_single(self, value):
        match_list = list(re.finditer(r"\${(\w+:)?([\w\.%_-]+?)}", value))
        if len(match_list) == 0:
            return value

        root = self._get_root()
        if len(match_list) == 1 and value == match_list[0].group(0):
            # simple interpolation, inherit type
            match = match_list[0]
            return Config._resolve_value(root, match.group(1), match.group(2))
        else:
            orig = value
            new = ''
            last_index = 0
            for match in match_list:
                new_val = Config._resolve_value(root, match.group(1), match.group(2))
                new += orig[last_index:match.start(0)] + str(new_val)
                last_index = match.end(0)

            new += orig[last_index:]
            return new

    @staticmethod
    def is_primitive_type(value):
        """
        Ensures this value type is of a valid type
        Throws ValueError otherwise
        :param value:
        :return:
        """
        from .listconfig import ListConfig
        from .dictconfig import DictConfig
        # None is valid
        if value is None:
            return True
        valid = [bool, int, str, float, DictConfig, ListConfig, BaseNode]
        if six.PY2:
            valid.append(unicode)

        return isinstance(value, tuple(valid))

    @staticmethod
    def _item_eq(c1, k1, c2, k2):
        v1 = c1.content[k1]
        v2 = c2.content[k2]
        if isinstance(v1, BaseNode):
            v1 = v1.value()
            if isinstance(v1, str):
                # noinspection PyProtectedMember
                v1 = c1._resolve_single(v1)
        if isinstance(v2, BaseNode):
            v2 = v2.value()
            if isinstance(v2, str):
                # noinspection PyProtectedMember
                v2 = c2._resolve_single(v2)

        if isinstance(v1, Config) and isinstance(v2, Config):
            if not Config._config_eq(v1, v2):
                return False
        return v1 == v2

    @staticmethod
    def _list_eq(l1, l2):
        from .listconfig import ListConfig
        assert isinstance(l1, ListConfig)
        assert isinstance(l2, ListConfig)
        if len(l1) != len(l2):
            return False
        for i in range(len(l1)):
            if not Config._item_eq(l1, i, l2, i):
                return False

        return True

    @staticmethod
    def _dict_conf_eq(d1, d2):
        from .dictconfig import DictConfig
        assert isinstance(d1, DictConfig)
        assert isinstance(d2, DictConfig)
        if len(d1) != len(d2):
            return False
        k1 = sorted(d1.keys())
        k2 = sorted(d2.keys())
        if k1 != k2:
            return False
        for k in k1:
            if not Config._item_eq(d1, k, d2, k):
                return False

        return True

    @staticmethod
    def _config_eq(c1, c2):
        from .listconfig import ListConfig
        from .dictconfig import DictConfig
        assert isinstance(c1, Config)
        assert isinstance(c2, Config)
        if isinstance(c1, DictConfig) and isinstance(c2, DictConfig):
            return DictConfig._dict_conf_eq(c1, c2)
        if isinstance(c1, ListConfig) and isinstance(c2, ListConfig):
            return Config._list_eq(c1, c2)
        # if type does not match objects are different
        return False



from .config import Config
from .nodes import BaseNode, UntypedNode
from .errors import ReadonlyConfigError
import copy


class DictConfig(Config):

    def __init__(self, content, parent=None):
        super(DictConfig, self).__init__()
        assert isinstance(content, dict)
        self.__dict__['content'] = {}
        self.__dict__['parent'] = parent
        for k, v in content.items():
            self.__setitem__(k, v)

    def __deepcopy__(self, memodict={}):
        res = DictConfig({})
        self._deepcopy_impl(res)
        return res

    def __setitem__(self, key, value):
        assert isinstance(key, str)
        if not isinstance(value, Config) and (isinstance(value, dict) or isinstance(value, list)):
            from omegaconf import OmegaConf
            value = OmegaConf.create(value, parent=self)
        if not Config.is_primitive_type(value):
            full_key = self.get_full_key(key)
            raise ValueError("key {}: {} is not a primitive type".format(full_key, type(value).__name__))
        if self._get_flag('freeze'):
            raise ReadonlyConfigError(self.get_full_key(key))
        if key not in self.content and self._get_flag('struct') is True:
            raise KeyError("Accessing unknown key in a struct : {}".format(self.get_full_key(key)))

        input_config_or_node = isinstance(value, (BaseNode, Config))
        if key in self:
            # BaseNode or Config, assign as is
            if input_config_or_node:
                self.__dict__['content'][key] = value
            else:
                # primitive input
                if isinstance(self.__dict__['content'][key], Config):
                    # primitive input replaces config nodes
                    self.__dict__['content'][key] = value
                else:
                    self.__dict__['content'][key].set_value(value)
        else:
            if input_config_or_node:
                self.__dict__['content'][key] = value
            else:
                self.__dict__['content'][key] = UntypedNode(value)

    # hide content while inspecting in debugger
    def __dir__(self):
        return self.content.keys()

    def __setattr__(self, key, value):
        """
        Allow assigning attributes to DictConfig
        :param key:
        :param value:
        :return:
        """
        self.__setitem__(key, value)

    def __getattr__(self, key):
        """
        Allow accessing dictionary values as attributes
        :param key:
        :return:
        """
        return self.get(key=key, default_value=None)

    def __getitem__(self, key):
        """
        Allow map style access
        :param key:
        :return:
        """
        return self.__getattr__(key)

    def get(self, key, default_value=None):
        return self._resolve_with_default(key=key, value=self.get_node(key), default_value=default_value)

    def get_node(self, key):
        value = self.__dict__['content'].get(key)
        if key not in self.content and self._get_flag('struct'):
            raise KeyError("Accessing unknown key in a struct : {}".format(self.get_full_key(key)))
        return value

    __marker = object()

    def pop(self, key, default=__marker):
        if self._get_flag('freeze'):
            raise ReadonlyConfigError(self.get_full_key(key))
        val = self.content.pop(key, default)
        if val is self.__marker:
            raise KeyError(key)
        return val

    def keys(self):
        return self.content.keys()

    def __iter__(self):
        return iter(self.keys())

    def items(self, resolve=True):
        class MyItems(object):
            def __init__(self, m):
                self.map = m
                self.iterator = iter(m)

            def __iter__(self):
                return self

            # Python 3 compatibility
            def __next__(self):
                return self.next()

            def next(self):
                k = next(self.iterator)
                if resolve:
                    v = self.map.get(k)
                else:
                    v = self.map.content[k]
                    if isinstance(v, BaseNode):
                        v = v.value()
                kv = (k, v)
                return kv

        return MyItems(self)

    def __eq__(self, other):
        if isinstance(other, dict):
            return Config._dict_conf_eq(self, DictConfig(other))
        if isinstance(other, DictConfig):
            return Config._dict_conf_eq(self, other)
        return NotImplemented

    def __ne__(self, other):
        x = self.__eq__(other)
        if x is not NotImplemented:
            return not x
        return NotImplemented

    def __hash__(self):
        # TODO: should actually iterate
        return hash(str(self))



"""OmegaConf module"""
import copy
import io
import os
import sys
import warnings
from collections import defaultdict
from .config import Config
import yaml


def register_default_resolvers():
    def env(key):
        try:
            return os.environ[key]
        except KeyError:
            raise KeyError("Environment variable '{}' not found".format(key))

    OmegaConf.register_resolver('env', env)


class OmegaConf:
    """OmegaConf primary class"""

    def __init__(self):
        raise NotImplementedError("Use one of the static construction functions")

    @staticmethod
    def create(obj=None, parent=None):
        from .dictconfig import DictConfig
        from .listconfig import ListConfig
        from .config import get_yaml_loader

        if isinstance(obj, str):
            new_obj = yaml.load(obj, Loader=get_yaml_loader())
            if new_obj is None:
                new_obj = {}
            elif isinstance(new_obj, str):
                new_obj = {obj: None}
            return OmegaConf.create(new_obj)
        else:
            if obj is None:
                obj = {}

            if isinstance(obj, dict):
                return DictConfig(obj, parent)
            elif isinstance(obj, (list, tuple)):
                return ListConfig(obj, parent)
            else:
                raise RuntimeError("Unsupported type {}".format(type(obj).__name__))

    @staticmethod
    def empty():
        warnings.warn("Use OmegaConf.create() (since 1.1.5)", DeprecationWarning,
                      stacklevel=2)
        """Creates an empty config"""
        return OmegaConf.create()

    @staticmethod
    def load(file_):
        from .config import get_yaml_loader
        if isinstance(file_, str):
            with io.open(os.path.abspath(file_), 'r', encoding='utf-8') as f:
                return OmegaConf.create(yaml.load(f, Loader=get_yaml_loader()))
        elif getattr(file_, 'read'):
            return OmegaConf.create(yaml.load(file_, Loader=get_yaml_loader()))
        else:
            raise TypeError("Unexpected file type")

    @staticmethod
    def from_filename(filename):
        warnings.warn("use OmegaConf.load() (since 1.1.5)", DeprecationWarning,
                      stacklevel=2)

        """Creates config from the content of the specified filename"""
        assert isinstance(filename, str)
        return OmegaConf.load(filename)

    @staticmethod
    def from_file(file_):
        """Creates config from the content of the specified file object"""
        warnings.warn("use OmegaConf.load() (since 1.1.5)", DeprecationWarning,
                      stacklevel=2)
        return OmegaConf.load(file_)

    @staticmethod
    def from_string(content):
        from .config import get_yaml_loader
        warnings.warn("use OmegaConf.create() (since 1.1.5)", DeprecationWarning,
                      stacklevel=2)
        """Creates config from the content of string"""
        assert isinstance(content, str)
        yamlstr = yaml.load(content, Loader=get_yaml_loader())
        return OmegaConf.create(yamlstr)

    @staticmethod
    def from_dict(dict_):
        """Creates config from a dictionary"""
        warnings.warn("use OmegaConf.create() (since 1.1.5)", DeprecationWarning,
                      stacklevel=2)
        assert isinstance(dict_, dict)
        return OmegaConf.create(dict_)

    @staticmethod
    def from_list(list_):
        """Creates config from a list"""
        warnings.warn("use OmegaConf.create() (since 1.1.5)", DeprecationWarning,
                      stacklevel=2)
        assert isinstance(list_, list)
        return OmegaConf.create(list_)

    @staticmethod
    def from_cli(args_list=None):
        if args_list is None:
            # Skip program name
            args_list = sys.argv[1:]
        return OmegaConf.from_dotlist(args_list)

    @staticmethod
    def from_dotlist(dotlist):
        """
        Creates config from the content sys.argv or from the specified args list of not None
        :param dotlist:
        :return:
        """
        conf = OmegaConf.create()
        conf.merge_with_dotlist(dotlist)
        return conf

    @staticmethod
    def merge(*others):
        """Merge a list of previously created configs into a single one"""
        assert len(others) > 0
        target = copy.deepcopy(others[0])
        target.merge_with(*others[1:])
        return target

    # noinspection PyProtectedMember
    @staticmethod
    def register_resolver(name, resolver):
        assert callable(resolver), "resolver must be callable"
        assert name not in Config._resolvers, "resolved {} is already registered".format(name)

        # noinspection PyProtectedMember
        def caching(config, key):
            cache = Config._resolvers_cache[id(config)][name]
            val = cache[key] if key in cache else resolver(key)
            cache[key] = val
            return val

        Config._resolvers[name] = caching

    # noinspection PyProtectedMember
    @staticmethod
    def get_resolver(name):
        return Config._resolvers[name] if name in Config._resolvers else None

    # noinspection PyProtectedMember
    @staticmethod
    def clear_resolvers():
        Config._resolvers = {}
        Config._resolvers_cache = defaultdict(lambda: defaultdict(dict))
        register_default_resolvers()

    @staticmethod
    def set_readonly(conf, value):
        # noinspection PyProtectedMember
        conf._set_flag('freeze', value)

    @staticmethod
    def is_readonly(conf):
        # noinspection PyProtectedMember
        return conf._get_flag('freeze')

    @staticmethod
    def set_struct(conf, value):
        # noinspection PyProtectedMember
        conf._set_flag('struct', value)

    @staticmethod
    def is_struct(conf):
        # noinspection PyProtectedMember
        return conf._get_flag('struct')


# register all default resolvers
register_default_resolvers()



"""
OmegaConf setup
    Instructions:
    # Build:
    rm -rf dist/ omegaconf.egg-info/
    python3 setup.py sdist bdist_wheel
    # Upload:
    python3 -m twine upload dist/*
"""
import setuptools

with open("README.md", "r") as fh:
    LONG_DESC = fh.read()
    setuptools.setup(
        name="omegaconf",
        version="1.3.0rc7",
        author="Omry Yadan",
        author_email="omry@yadan.net",
        description="A flexible configuration library",
        long_description=LONG_DESC,
        long_description_content_type="text/markdown",
        setup_requires=["pytest-runner"],
        tests_require=["pytest"],
        url="https://github.com/omry/omegaconf",
        keywords='yaml configuration config',
        packages=['omegaconf'],
        classifiers=[
            "Programming Language :: Python :: 2.7",
            "Programming Language :: Python :: 3.5",
            "Programming Language :: Python :: 3.6",
            "Programming Language :: Python :: 3.7",
            "License :: OSI Approved :: BSD License",
            "Operating System :: OS Independent",
        ],
        install_requires=['six', 'PyYAML']
    )
